// ftpx.cpp - written by vesely in milano on 27 jan 1998
// perform ftp file transfer based on file's date and time
//
// Copyright (c) 1998 Alessandro Vesely
// see NOTICE.txt


#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <strstream>
#include <vector>
#include <list>

#include <windows.h>
#include <wininet.h>

#include <assert.h>
#include <direct.h>

// for isatty(fileno(stdout)) in main
#include <stdio.h>
#include <io.h>

using namespace std;

// usage messages and global variables
int verbose = 0; // set while reading args;
const char help_msg[] =
"ftpx [-[V[V]]h] |\n"
"  [@script] [-Lpath] [-Rpath] [-Hhost] [-Uuser] [-Ppass] [-Tbias]\n"
"\t[[-20pgntabisrV] [-]file]...\n";

const char copyright[] =
"ftpx V1.1 - see <http://www.tana.it/sw/ftpx.html>\n"
"Copyright (c) 1998 Alessandro Vesely\n";

const char short_help_msg[] =
"Localpath Remotepath Host User [Password] Timebias [VVVery] Verbose\n"
"-? this help, -h a longer one\n"
"per file options:\n"
"\t2 ways | 0 ways | put | get\n"
"\tno touch | touch after put\n"
"\tascii | binary\n"
"\tignore case | sensitive | rename local file\n";

const char long_help_msg[] =
"where:\n"
"file\tcan be given using wildcards. File list xyz can be given as @xyz,\n"
"\tthe content of xyz is the same as the arguments, except a line\n"
"\nstarting with # is a comment. File '-abc' must be given as '--abc'.\n"
"\nFile '@abc' not meant to be a list of files or options must be given\n"
"\tas '-@abc'.\n"
"\n"
"Lower case options affect all files that follow them. They are divided\n"
"in three groups, any option resetting the other(s) in the same group.\n"
"20pg\tfor 2 ways, none or only put or get. In any case, puts or gets\n"
"\twill only overwrite the older file."
"\tThe default is 2 way, i.e. either get or put depending on date/time.\n"
"\t0 is for listing, renaming or redating local files.\n"
"\n"
"nt\t[no] touch files after they've been put, default is no-touch\n"
"\t(Files are touched after get anyway.)\n"
"\n"
"ab\tfor ascii or binary. The default is ascii.\n"
"\n"
"isr\tfor case: ignore, sensitive or rename. Blanks in filenames are\n"
"\treplaced with underscore and back slashes with forward ones. The\n"
"\tdefault is to ignore case (tolower for new files). The rename option\n"
"\twill rename the local file.\n"
"\n"
"Uppercase options affect the whole run:\n"
"L\tis the local directory path (the current directory by default)\n"
"R\tis the remote directory path (the current directory by default)\n"
"T\tis the remote bias adjustment from the remote system, in minutes\n"
"\t(remote time = local time + bias). It defaults to 0\n"
"H\tis the remote host\n"
"U\tthe remote user. H and U have no default, if neither is specified\n"
"\tthe program will still list and optionally rename (to lowercase)\n"
"\tthe local files.\n"
"LRHUT must be immediately followed by an argument, e.g. -LC:\MYDIR.\n"
"It is possible to specify a current directory for each drive, they will\n"
"then be referred to by relative filespecs such as C:MYFILE, but they\n"
"will all match a unique remote current directory.\n"
"\n"
"V\tis for verbose. Specifying it multiple times results in debug output.\n";

///////////////////////////////////////////////////////////
// class teebuf (see Pete Becker's column on sept 1997 CUJ)

#if defined( __TURBOC__)
#define xsputn do_sputn
typedef int streamsize;
typedef int int_type;
typedef char char_type;
#endif

static const int_type traits_eof = EOF; // traits::eof() ?

class teebuf : public streambuf
{
   streambuf *buf1, *buf2;

public:
   explicit teebuf(streambuf *s1 = 0, streambuf *s2 = 0) :
      buf1(s1), buf2(s2) {}
   streamsize xsputn(const char_type *, streamsize);
   int_type overflow(int_type);
};

class teestream : public ostream
{
   teebuf buf;

public:
   teestream(ostream&, ostream&);
};

//

streamsize teebuf::xsputn(const char_type *s, streamsize n)
{
   streamsize rtc1 = buf1 ? buf1->sputn(s, n) : 0;
   streamsize rtc2 = buf2 ? buf2->sputn(s, n) : 0;
   return ((int)rtc1 > (int)rtc2) ? rtc1 : rtc2;
}

int_type teebuf::overflow(int_type c)
{
   int_type rtc1 = buf1 ? buf1->sputc(c) : traits_eof;
   int_type rtc2 = buf2 ? buf2->sputc(c) : traits_eof;
   return rtc1 == rtc2 && rtc1 == traits_eof ? traits_eof : c;
}

teestream::teestream(ostream& o1, ostream& o2) :
  ostream(&buf), buf(o1.rdbuf(), o2.rdbuf())
{
}


// static errbuf for cerr (assigned in main)
streambuf *get_errbuf(bool and_cout)
{
   static teebuf errbuf(cerr.rdbuf(), and_cout ? cout.rdbuf() : 0);

   return &errbuf;
}


#if defined(xsputn)
#undef xsputn
#endif

//////////////////////////////////////////////////////////////////////////
// general utility functions


void do_report_error(const char *msg,
   const char *msg1, const char *msg2, const char *msg3,
   DWORD err, const char *errmsg, ostream& o)
{
   if (msg != 0) o << msg;
   if (msg1 != 0) o << msg1;
   if (msg2 != 0) o << msg2;
   if (msg3 != 0) o << msg3;
   o << ": ";
   if (err != 0) o << err << " ";
   o << errmsg << endl;
}

void do_error(const char *msg,
   const char *msg1, const char *msg2, const char *msg3,
   DWORD err, const char *buf)
{
   do_report_error(msg, msg1, msg2, msg3, err, buf, cerr);
}

void report_error(const char *msg = 0,
   const char *msg1 = 0, const char *msg2 = 0, const char *msg3 = 0)
{
   DWORD last_error = GetLastError();
   if (last_error == ERROR_INTERNET_EXTENDED_ERROR)
   {
      char buf[1024];
      DWORD err, siz = sizeof buf;

      if (InternetGetLastResponseInfo(&err, buf, &siz))
      {
         do_error(msg, msg1, msg2, msg3, err, buf);
         return;
      }
   }
   LPVOID lpMsgBuf = 0;
   FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      last_error,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
      (LPTSTR) &lpMsgBuf,
      0,
      NULL );

   char *pmsg = (char *) lpMsgBuf;
   do_error(msg, msg1, msg2, msg3, 0, pmsg);

   // Free the buffer.
   if ( lpMsgBuf != 0)
      LocalFree( lpMsgBuf );
}

ostream& operator<<(ostream& o, const FILETIME& ft)
{
   long fl = o.flags();
   SYSTEMTIME st;
   FileTimeToSystemTime(&ft, &st);
   o.setf(ios::right, ios::adjustfield);
   char fi = o.fill('0');
   o << setw(2) << st.wDay      << setw(0) << '/'
      << setw(2) << st.wMonth    << setw(0) << '/'
      << setw(2) << st.wYear     << setw(0) << ' '
      << setw(2) << st.wHour     << setw(0) << ':'
      << setw(2) << st.wMinute   << setw(0) << ':'
      << setw(2) << st.wSecond   << setw(0);
   o.fill(fi);
   o.flags(fl);
   return o;
}

char bool_YN(bool b)
{
   return b ? 'Y' : 'N';
}

bool get_drive(char& drive)
{
   int curr = _getdrive();
   bool ok = false;
   if (drive == 0)
   {
      drive = 'A' - 1 + curr;
      ok = true;
   }
   else
   {
      if (islower(drive)) drive += 'A' - 'a';
      if (isupper(drive))
      {
         int n = drive - 'A' + 1;
         ok = _chdrive(n) == 0;
         _chdrive(curr);
      }
      if (!ok)
      {
         cerr << "Invalid drive " << drive << ':' << endl;
      }
   }
   return ok;
}

int set_file_time(const char *fname, const FILETIME *create_ft)
{
   FILETIME ft;
   if (LocalFileTimeToFileTime(create_ft, &ft) == FALSE)
   {
      report_error("Getting FileTime");
      return 1;
   }
   HANDLE h = CreateFile(fname, GENERIC_READ | GENERIC_WRITE,
      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
   if (h == INVALID_HANDLE_VALUE)
   {
      report_error("Setting date of ", fname);
      return 1;
   }
   int rtc;
   if (SetFileTime(h, 0, 0, &ft) == FALSE)
   {
      report_error("Changing date of ", fname);
      rtc = 1;
   }
   else rtc = 0;
   CloseHandle(h);
   return rtc;
}

// A quick little routine that will center a window on the screen.
// Handy for dialog boxes
BOOL CenterWindow (HWND hwnd)
{
    RECT    rect;
    int     w, h;
    int     wScreen, hScreen, xNew, yNew;
    HDC     hdc;

    GetWindowRect (hwnd, &rect);
    w = rect.right - rect.left;
    h = rect.bottom - rect.top;

    hdc = GetDC (hwnd);
    wScreen = GetDeviceCaps (hdc, HORZRES);
    hScreen = GetDeviceCaps (hdc, VERTRES);
    ReleaseDC (hwnd, hdc);

    xNew = wScreen/2 - w/2;
    yNew = hScreen/2 - h/2;

    return SetWindowPos (hwnd, NULL, xNew, yNew, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

bool getdlgstring(unsigned cid, string& s, HWND hdlg)
{
   HWND hctrl = GetDlgItem(hdlg, cid);
   WORD l = (WORD) SendMessage(hctrl, EM_LINELENGTH, 0, 0);
   char buf[64];
   if (l >= sizeof buf || l <= 0)
   {
      SetFocus(hctrl);
      SendMessage(hctrl, EM_SETSEL, 0, -1);
      MessageBox(hdlg, l ? "Value too long" : "Missing value", "Error", MB_OK);
      return false;
   }
   *((LPWORD) buf) = l;
   SendMessage(hctrl, EM_GETLINE, 0, (LPARAM)buf);
   buf[l] = 0;
   s = buf;
   return true;
}


void remove_space(string& r)
{
   for(unsigned int i = 0; i < r.size(); ++i)
      if (isspace(r[i]))
         r[i] = '_';
}

void lower_nospace(string& r)
{
   for(unsigned int i = 0; i < r.size(); ++i)
   {
      char c = r[i];
      if (isspace(c))
         r[i] = '_';
      else if (isupper(c))
         r[i] = tolower(c);
   }
}

int wildstr(const char *card, const char *name)
{
   assert(card != 0 && name != 0);

   unsigned char c = *card, n = *name;
   while(c != 0 && n != 0)
      if (c == n || c == '?')
      {
         c = *++card;
         n = *++name;
      }
      else if (c == '*')
      {
         c = card[1];
         while (n != c)
            if ((n = *++name) == 0) return c == 0 ? 0 : 1;
         // matching char found? if rest doesn't match then skip it
         if (wildstr(card + 1, name) == 0) return 0;
         c = '*';
         n = *++name;
      }
      else return 1;
   return c == n || (c == '*' && card[1] == 0) ? 0 : 1;
}

int wildistr(const char *card, const char *name)
{
   assert(card != 0 && name != 0);

   unsigned char c = *card, n = *name;
   while(c != 0 && n != 0)
      if (c == n || c == '?' || tolower(c) == tolower(n))
      {
         c = *++card;
         n = *++name;
      }
      else if (c == '*')
      {
         c = tolower(card[1]);
         while (n != c && tolower(n) != c)
            if ((n = *++name) == 0) return c == 0 ? 0 : 1;
         // matching char found? if rest doesn't match then skip it
         if (wildistr(card + 1, name) == 0) return 0;
         c = '*';
         n = *++name;
      }
      else return 1;
   assert(c == 0 || n == 0);
   return c == n || (c == '*' && card[1] == 0) ? 0 : 1;
}

/////////////////////////////////////////////////////////////////////////
// specific classes (implemented near the end of file)

class flag_set
{
   unsigned flag;

public:
   flag_set() : flag(0) {}

   enum flag_value { l_done = 1, r_done = 2, fail = 4, l_match = 8,
      r_match = 16, root = 32, current = 64, further = 128};

   bool is_local_done() const { return (flag & l_done) != 0; }
   bool is_remote_done() const { return (flag & r_done) != 0; }
   bool is_fail() const { return (flag & fail) != 0; }
   bool is_local_match() const { return (flag & l_match) != 0; }
   bool is_remote_match() const { return (flag & r_match) != 0; }
   bool is_root() const { return (flag & root) != 0; }
   bool is_current() const { return (flag & current) != 0; }
   bool is_further() const { return (flag & further) != 0; }

   void set_local_done() { flag |= l_done; }
   void set_remote_done() { flag |= r_done; }
   void set_fail() { flag |= fail; }
   void set_local_match() { flag |= l_match; }
   void set_remote_match() { flag |= r_match; }
   void set_root() { flag |= root; }
   void set_current() { flag |= current; }
   void set_further() { flag |= further; }
};

class file_opt
{
   unsigned char xfer_flag, mode_flag, upper_flag, flag;

public:
   enum xfer { no_way, put, get, two_way};
   enum mode { ascii, binary};
   enum upper { sensitive, ignore, rename};
   enum etc { minus = 1, wild = 2, date_set = 4};

   void assert_values() const
    { assert(xfer_flag <= two_way &&
         mode_flag <= binary && upper_flag <= rename); }

   file_opt(xfer x = no_way, mode m = ascii, upper u = sensitive) :
      xfer_flag(x), mode_flag(m), upper_flag(u), flag(0) { assert_values(); }

   bool is_put() const { return (xfer_flag & put) != 0; }
   bool is_get() const { return (xfer_flag & get) != 0; }
   void set_xfer(xfer x) { xfer_flag = x; assert_values(); }

   bool is_ascii() const { return mode_flag == ascii; }
   bool is_binary() const { return mode_flag == binary; }
   void set_mode(mode m) { mode_flag = m; assert_values(); }

   bool is_ignore() const { return upper_flag != sensitive; }
   bool is_rename() const { return upper_flag == rename; }
   bool is_sensitive() const { return upper_flag == sensitive; }
   void set_upper(upper u) { upper_flag = u; assert_values(); }

   bool is_minus() const { return (flag & minus) != 0; }
   void reset_minus() { flag &= ~minus; }
   void set_minus() { flag |= minus; }

   bool is_wild() const { return (flag & wild) != 0; }
   void reset_wild() { flag &= ~wild; }
   void set_wild() { flag |= wild; }

   bool is_date_set() const { return (flag & date_set) != 0; }
   void reset_date_set() { flag &= ~date_set; }
   void set_date_set() { flag |= date_set; }
};

struct directory : public flag_set
{
   string local, remote;
   directory *prev;
   char drive; // only if is root or current

   directory() : local(), remote(), prev(0), drive(0) {}

   enum path_mode { local_path, remote_path};
   int build_path(string& dest, path_mode pm = remote_path) const;
};

bool operator<(const directory& d1, const directory& d2)
{
   const directory *dd1 = d1.prev, *dd2 = d2.prev;
   if (dd1 == dd2)
      return d1.remote < d2.remote;
   else if (dd1 == 0 || dd2 == 0)
      return dd1 == 0;
   else
      return *dd1 < *dd2;
}

bool operator==(const directory& d1, const directory& d2)
{
   const directory *dd1 = d1.prev, *dd2 = d2.prev;
   return dd1 == dd2 &&
      (dd1 == 0 || *dd1 == *dd2) &&
      d1.is_fail() == d2.is_fail() &&
      d1.is_root() == d2.is_root() &&
      d1.remote == d2.remote;
}

class dir_v : public list<directory>
{
   int nroot, ncurrent;

public:
   dir_v() : nroot(0), ncurrent(0) {}
   directory *current_directory(const char *path = 0, char drive = 0);
   directory *root_directory(char drive = 0);
   directory *find_directory(const char *name,       // dir name to find
      const directory *prev,                         // parent directory
      const string directory::*member);              // local or remote
};

struct file : public flag_set
{
   directory *path;
   string local, remote;
   file_opt opt;
   FILETIME remote_ft, local_ft;
   long remote_size, local_size;

   file() : path(0), remote_size(0), local_size(0)
      { local_ft.dwLowDateTime = local_ft.dwHighDateTime =
           remote_ft.dwLowDateTime = remote_ft.dwHighDateTime = 0; }
   void dump() const;
};

bool operator<(const file& f1, const file& f2)
{
   const directory *d1 = f1.path, *d2 = f2.path;
   if (d1 == d2)
      return f1.remote < f2.remote;
   else if (d1 == 0 || d2 == 0)
      return d1 == 0;
   else
      return *d1 < *d2;
}

bool operator==(const file& f1, const file& f2)
{
   return f1.path == f2.path && f1.remote == f2.remote && f1.local == f2.local;
}

class file_v : public list<file>
{
public:
   iterator find_file(const char *name,
      const directory *path,
      const string file::*member);
};

class ftp_link
{
   string m_directory, m_user, m_host, m_pass;
   string CWD;
   HINTERNET hopen, hconn;
   TIME_ZONE_INFORMATION tz_info;
   FILETIME future_time, patch6_start, patch6_end;
   bool connected;

public:
   enum mode {ascii, binary};
   ftp_link() : hopen(0), hconn(0), connected(false)
      { memset(&tz_info, 0, sizeof tz_info);

        SYSTEMTIME st, st1;
        GetLocalTime(&st);

        // build gen, 1st next year = end of patch period
        // (see patch6 comment in FileTimeToLocalFileTime)
        st1.wDayOfWeek = 0; // ??
        st1.wYear = st.wYear + 1;
        st1.wMonth = 1;
        st1.wDay = 1;
        st1.wHour = 0;
        st1.wMinute = 0;
        st1.wSecond = 0;
        st1.wMilliseconds = 0;
        if (SystemTimeToFileTime(&st1, &patch6_end) == FALSE)
           memset(&patch6_end, 0, sizeof patch6_end);

        // build 6 months from now = start of patch period
        memcpy(&st1, &st, sizeof st1);
        if (st.wMonth > 6)
        {
           st1.wMonth -= 6;
           st1.wYear += 1;
        }
        else
           st1.wMonth += 6;
        if (SystemTimeToFileTime(&st1, &patch6_start) == FALSE)
           memset(&patch6_start, 0, sizeof patch6_start);
        else
           patch6_start.dwHighDateTime -= 201; // that's about 23.98 hours

        // build future time = approx 1 day from now
        if (SystemTimeToFileTime(&st, &future_time) == FALSE)
           memset(&future_time, 0, sizeof future_time);
        else
           future_time.dwHighDateTime += 201; // that's about 23.98 hours
      }
   ~ftp_link() { if (is_connect()) disconnect(); }

   // do
   bool connect(); // true if successful
   bool get_current_directory(string&);
   bool set_working_directory(const string&);
   bool put_file(const string& local, const string& remote, mode m);
   bool get_file(const string& local, const string& remote, mode m);
   bool FileTimeToLocalFileTime(const FILETIME*a, FILETIME*b,
     /* file name for error reporting */ const char*fname = 0)
    { // FtpFindFirstFile delivers the date "AS IS", i.e. local time
      if (tz_info.Bias == 0)
         memcpy(b, a, sizeof(FILETIME));
      else
       { SYSTEMTIME st, st_local;
         if (FileTimeToSystemTime(a, &st) == FALSE ||
            SystemTimeToTzSpecificLocalTime(&tz_info, &st, &st_local) == FALSE ||
            SystemTimeToFileTime(&st_local, b) == FALSE)
             { report_error("converting time"); return false; }}

      // check patch6:
      // some system (e.g.Solaris) report a date such as
      // "nov 2 17:30" until may 1st 1998 at 18:30 and
      // "nov 2  1997" thereafter. Some version of wininet
      // fill in the current year in the former case, which is wrong.
      // To work around that, refuse any future date, except those
      // that are within the patch period. A 23.98 hours slack time
      // is provided for in c'tor to take possible sync problems
      // out of the way.

      if (CompareFileTime(&future_time, b) <= 0)
         if (CompareFileTime(&patch6_start, b) <= 0 &&
          CompareFileTime(&patch6_end, b) > 0)
          { SYSTEMTIME st;
            if (FileTimeToSystemTime(b, &st) == FALSE ||
               (st.wYear -= 1, SystemTimeToFileTime(&st, b)) == FALSE)
                { report_error("converting patch6 time"); return false; }}
         else
          { cerr << "Converted date " << *b;
            if (fname) cerr << " for " << fname;
            cerr << " is greater than " << future_time << " and not patch6!"
               << endl;
            return false; }
      return true;
    }
   bool disconnect();

   // get
   bool is_connect() const { return connected; }
   bool is_named() const { return m_host.size() > 0; }
   HINTERNET get_handle() const { assert(is_connect()); return hconn; }

   // set, return 0 if ok
   int current_directory(const char *arg)
      { return set_if(m_directory, arg, "remote directory"); }
   int host(const char *arg) { return set_if(m_host, arg, "host"); }
   int user(const char *arg) { return set_if(m_user, arg, "user"); }
   int password(const char *arg) { return set_if(m_pass, arg, "password"); }
   long set_time_bias(long bias)
    { long tmp = tz_info.Bias; tz_info.Bias = bias; return tmp; }

private:
   int set_if(string& s, const char *arg, const char *def)
    { if (arg == 0 || *arg == 0 || s.size() > 0)
       { cerr << "Invalid " << def << " \"" << arg << "\"" << endl;
         return 1; }
      s = arg; return 0; }
};

////////////////////////////////////////////////////////////////////
// specific functions

int match_remote(directory *base_dir, dir_v& dirv, file_v& filev, ftp_link& ftp)
{
   assert(base_dir);

   string path;
   if (base_dir->build_path(path, directory::remote_path) != 0)
      return 1;   // path check and error reporting

   vector<directory*> match;
   match.push_back(base_dir);
   for (dir_v::iterator i = dirv.begin(); i != dirv.end(); ++i)
      if (&*i != base_dir && *i == *base_dir)
         match.push_back(&*i);

   // assume remote directory is already set
   int rtc = 1;
   WIN32_FIND_DATA find;
   HINTERNET hsearch = FtpFindFirstFile(ftp.get_handle(), 0, &find,
      INTERNET_FLAG_DONT_CACHE, 0);
   if (hsearch)
   {
      do
      {
         for (vector<directory*>::iterator i = match.begin();
            i != match.end(); ++i)
         {
            directory *dir = *i;
            directory *cur = dirv.find_directory(find.cFileName, dir,
               &directory::remote);
            if (cur != 0)
            {
#if defined(CAN_TRACK_REMOTE_LINKS)
               // if path is a link to a remote directory,
               // it will be flagged as a plain file
               if ((find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
               {
                  cerr << "Remote directory " << path << find.cFileName
                     << " is a plain file" << endl;
                  cur->set_fail();
               }
               else
#endif // CAN_TRACK_REMOTE_LINKS
               if (cur->is_remote_match())
               {
                  cerr << "Ambiguous directories in " << path << ": "
                     << cur->remote << " and " << find.cFileName << endl;
                  cur->set_fail();
               }
               else
               {
                  cur->remote = find.cFileName;
                  cur->set_remote_match();
                  if (verbose > 3)
                     cout << "Matched remote directory "
                        << path << find.cFileName << endl;
               }
            }

            file_v::iterator fit =
               filev.find_file(find.cFileName, dir, &file::remote);
            file *f = fit != filev.end() ? &*fit : 0;
            if (f != 0 && !f->opt.is_wild()) // match an existing file
            {
               if ((find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
               {
                  cerr << "Remote file " << path << find.cFileName
                     << " is a directory" << endl;
                  f->set_fail();
                  f = 0;
               }
               else if (f->is_remote_match())
               {
                  cerr << "Ambiguous files in " << path << ": "
                     << f->remote << " and " << find.cFileName << endl;
                  f->set_fail();
                  f = 0;
               }
               else if (verbose > 3)
                  cout << "Matched remote file "
                     << path << find.cFileName << endl;
            }
            else if (f != 0 && // match a wild card: new file
               (find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
            {
               // check a local file doesn't already exist with this name
               if (f->opt.is_sensitive())
               {
                  file_v::iterator ckfit =
                     filev.find_file(find.cFileName, dir, &file::local);
                  if (ckfit != filev.end() && !(*ckfit).opt.is_wild())
                  {
                     cerr << "Case conflict: " << (*ckfit).remote
                        << " and " << find.cFileName
                        << " in remote directory " << path << endl;
                     (*ckfit).set_fail();
                     f = 0;
                  }
               }

               if (f != 0)
               {
                  file nf;
                  nf.path = dir;
                  nf.opt = f->opt;
                  nf.opt.reset_wild();
                  // nf.local set below (if std::string keeps copies)
                  f = &*filev.insert(fit, nf);
                  if (verbose > 2)
                     cout << "Found remote file "
                        << path << find.cFileName << endl;
               }
            }
            else f = 0;

            if (f != 0)
            {
               f->remote = find.cFileName;
               if (f->local.empty()) f->local = f->remote; // new files
               f->set_remote_match();
               f->remote_size = find.nFileSizeLow;
               if (!ftp.FileTimeToLocalFileTime
                  (&find.ftLastWriteTime, &f->remote_ft, f->remote.c_str()))
                     f->set_fail();
               else if (find.nFileSizeHigh != 0)
               {
                  cerr << "Remote file " << path << f->remote
                     << " is too large" << endl;
                  f->set_fail();
               }
            }
         }
      } while (InternetFindNextFile(hsearch, &find));

      if (rtc >= 0)
      {
         if (GetLastError() == ERROR_NO_MORE_FILES)
            rtc = 0;
         else report_error("Getting file list in ", path.c_str());
      }
      else rtc = 1;

      if (!InternetCloseHandle(hsearch))
      {
         report_error("Closing ls in ", path.c_str());
         rtc = 1;
      }
   }
   else if (GetLastError() == ERROR_NO_MORE_FILES)
      rtc = 0;
   else report_error("Starting file list in ", path.c_str());

   if (rtc == 0)
      for (vector<directory*>::iterator i = match.begin();
         i != match.end(); ++i)
            if (*i != base_dir) // set by caller
               (*i)->set_remote_done();

   return rtc;
}

int set_remote_directory(file_opt& fopt, directory *dir,
   dir_v& dirv, file_v& filev, ftp_link& ftp)
{
   assert(dir != 0);
   assert(ftp.is_connect());

   string path;
   if (dir->build_path(path, directory::remote_path) != 0)
      return 1;

   if (verbose > 2)
      cout << "trying remote directory " << path << endl;

   if (ftp.set_working_directory(path))
      return 0;  // Ok

   if (dir->is_remote_match())
   {
      report_error("Cannot CWD to known path ", path.c_str());
      return 1;
   }

   char szBuffer[128], *szAltBuffer = 0;
   DWORD dwError, dwBufferLength = sizeof szBuffer;

   // perhaps not set because of mismatched case
   directory *prev = dir->prev;
   if (prev != 0 &&
      !prev->is_remote_done() &&
      !prev->is_fail() &&
      fopt.is_ignore() &&
      GetLastError() == ERROR_INTERNET_EXTENDED_ERROR &&
      (InternetGetLastResponseInfo(&dwError, szBuffer, &dwBufferLength) ||
         GetLastError() == ERROR_INSUFFICIENT_BUFFER &&
         (szAltBuffer = (char*)malloc(dwBufferLength)) != 0 &&
         InternetGetLastResponseInfo(&dwError, szAltBuffer, &dwBufferLength)))
   {
      const char *msg = (szAltBuffer != 0) ? szAltBuffer : szBuffer;
      int ftp_error = atoi(msg);
      if (ftp_error != 550 && ftp_error != 500 && ftp_error != 504)
      {
         cerr << "Invalid directory " << path << ": "
            << msg << " [Error " << int(dwError) << ", " << ftp_error
            << "]" << endl;
         if (szAltBuffer != 0) free(szAltBuffer);
         // actually any other 500+ result codes could lead to a retry
         // except "530 Not logged in", e.g.:
         // "500 Syntax error, command unrecognized"
         // "501 Syntax error in parameters or arguments"
         // "504 Command not implemented for that parameter"
         // "550 requested action not taken. File unavailable"
         // "553 Requested action not taken. File name not allowed."
         // ... need to check more servers ???
         dir->set_fail();
         return 1;
      }
      if (verbose > 1)
         cout << "ftp: Cannot CWD to " << path << ": "
            << msg << " [Error " << int(dwError) << ", " << ftp_error
            << "]\nTrying parent directory..." << endl;

      if (szAltBuffer != 0) free(szAltBuffer);

      if (set_remote_directory(fopt, prev, dirv, filev, ftp) == 0 &&
         match_remote(prev, dirv, filev, ftp) == 0)
      {
         prev->set_remote_done();
         if (dir->build_path(path, directory::remote_path) == 0 &&
            ftp.set_working_directory(path))
               return 0;
      }
   }

   report_error("Invalid directory ", path.c_str());
   dir->set_fail();
   return 1;
}

int add_remote_files(dir_v& dirv, file_v& filev, ftp_link& ftp)
{
   assert(ftp.is_connect());

   for (file_v::iterator i = filev.begin(); i != filev.end(); ++i)
   {
      file& f = *i;
      directory *dir = f.path;
      if (dir == 0)
      {
         cerr << "Internal error 2: " << f.local << " has no path" << endl;
         return 1;
      }

      if (!dir->is_remote_done())
      {
         if (set_remote_directory(f.opt, dir, dirv, filev, ftp) == 0 &&
            match_remote(dir, dirv, filev, ftp) == 0)
               dir->set_remote_done();
      }
   }
   return 0;
}

directory *track_remote(file_v::iterator redo, int nredo, directory *dir,
   string& lpath, string& rpath, file_v& filev, ftp_link& ftp)
{
   assert(redo == filev.end() || ftp.is_connect());
   assert(nredo >= 0);
   assert(nredo == 0 || redo != filev.end());

   if (redo != filev.end()) // list remote files to adjust date/time
   {
      WIN32_FIND_DATA find;
      HINTERNET hsearch = FtpFindFirstFile(ftp.get_handle(), 0, &find,
         INTERNET_FLAG_DONT_CACHE, 0);
      if (hsearch)
      {
         directory *oldpath = (*redo).path;
         do
         {
            for (file_v::iterator i = redo; i != filev.end(); ++i)
            {
               file& f = *i;
               if (f.path != oldpath)
                  break;
               if (f.is_remote_done() &&
                  strcmp(f.remote.c_str(), find.cFileName) == 0)
               {
                  f.remote_size = find.nFileSizeLow;
                  if (ftp.FileTimeToLocalFileTime
                     (&find.ftLastWriteTime, &f.remote_ft, f.remote.c_str()))
                  {
                     --nredo;
                     f.set_local_done();
                  }
                  break;
               }
            }
         } while (InternetFindNextFile(hsearch, &find));

         if (GetLastError() != ERROR_NO_MORE_FILES)
            report_error("Getting redo file list");

         if (!InternetCloseHandle(hsearch))
            report_error("Closing redo ls");
      }
      else if (GetLastError() != ERROR_NO_MORE_FILES)
         report_error("Starting redo file list");
   }

   if (nredo != 0)
   {
      cerr << "Number of files written to remote " << rpath
         << " differs by " << nredo << " files after new list" << endl;
   }

   if (dir == 0) return 0;

   // set remote path to this directory
   directory *cur = dir;

   if (cur->build_path(lpath, directory::local_path) != 0)
   {
      if (verbose > 1 && !cur->is_fail())
         cout << "Skipping local \"" << cur->local << "\" = remote \""
            << cur->remote << "\"" << endl;
      cur->set_fail();
      return 0;
   }

   if (ftp.is_connect())
   {

      if (cur->build_path(rpath, directory::remote_path) != 0)
      {
         if (verbose > 1 && !cur->is_fail())
            cout << "Skipping remote \"" << cur->remote << "\" = local \""
               << cur->local << "\"" << endl;
         cur->set_fail();
         return 0;
      }

      if (!cur->is_remote_done())
      {
         cerr << "Internal error: directory " << rpath
            << "was not done" << endl;
         cur->set_fail();
         return 0;
      }

      if (verbose > 3)
         cout << "Remote CWD to directory " << rpath << endl;

      if (!ftp.set_working_directory(rpath))
      {
         report_error("Cannot CWD to done path ", rpath.c_str());
         cur->set_fail();
         return 0;
      }
   }
   return cur;
}

int match_local(directory *dir, dir_v& dirv, file_v& filev)
{
   assert(dir);

   string path;
   if (dir->build_path(path, directory::local_path) != 0)
      return 1;

   int rtc = 1;
   path.append(1, '*');
   WIN32_FIND_DATA find;
   HANDLE hsearch = FindFirstFile(path.c_str(), &find);
   path.remove(path.size() - 1);
   if (hsearch != INVALID_HANDLE_VALUE)
   {
      do
      {
         directory *cur = dirv.find_directory(find.cFileName, dir,
            &directory::local);
         if (cur != 0)
         {
            if ((find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
            {
               cerr << "Local directory " << path << find.cFileName
                  << " is a plain file" << endl;
               cur->set_fail();
            }
            else if (cur->is_local_match())
            {
               cerr << "Ambiguous directories in " << path << ": "
                  << cur->local << " and " << find.cFileName << endl;
               cur->set_fail();
            }
            else
            {
               cur->local = find.cFileName;
               cur->set_local_match();
               if (verbose > 3)
                  cout << "Matched local directory "
                     << path << find.cFileName << endl;
            }
         }

         file_v::iterator fit =
            filev.find_file(find.cFileName, dir, &file::local);
         file *f = fit != filev.end() ? &*fit : 0;
         if (f != 0 && !f->opt.is_wild()) // match an existing file
         {
            if ((find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
            {
               cerr << "Local file " << path << find.cFileName
                  << " is a directory" << endl;
               f->set_fail();
               f = 0;
            }
            else if (f->is_local_match())
            {
               cerr << "Ambiguous files in " << path << ": "
                  << f->local << " and " << find.cFileName << endl;
               f->set_fail();
               f = 0;
            }
            else
            {
               // canonical name, unless different for sensitive comparison
               if (!f->opt.is_sensitive())
                  f->local = find.cFileName;
               if (verbose > 3)
                  cout << "Matched local file "
                     << path << find.cFileName << endl;
            }
         }
         else if (f != 0 && // match a wild card: new file
            (find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
         {
            file nf;
            nf.path = dir;
            nf.opt = f->opt;
            nf.opt.reset_wild();
            nf.remote = nf.local = find.cFileName;
            if (nf.opt.is_sensitive())
               remove_space(nf.remote);
            else // rename with lower, unless remote file exists
               lower_nospace(nf.remote);
            f = &*filev.insert(fit, nf);
            if (verbose > 2)
               cout << "Found local file "
                  << path << find.cFileName << endl;
         }
         else f = 0;

         if (f != 0)
         {
            f->set_local_match();
            f->local_size = find.nFileSizeLow;
            FileTimeToLocalFileTime(&find.ftLastWriteTime, &f->local_ft);
            if (find.nFileSizeHigh != 0)
            {
               cerr << "Local file " << path << f->local
                  << " is too large" << endl;
               f->set_fail();
            }
         }
      } while (FindNextFile(hsearch, &find));

      if (rtc >= 0)
      {
         if (GetLastError() == ERROR_NO_MORE_FILES)
            rtc = 0;
         else report_error("Getting file dir in ", path.c_str());
      }
      else rtc = 1;

      if (!FindClose(hsearch))
      {
         report_error("Closing dir in ", path.c_str());
         rtc = 1;
      }
   }
   else if (GetLastError() == ERROR_NO_MORE_FILES)
      rtc = 0;
   else report_error("Starting file dir in ", path.c_str());
   return rtc;
}


int add_local_files(dir_v& dirv, file_v& filev)
{
   for (file_v::iterator i = filev.begin(); i != filev.end(); ++i)
   {
      file& f = *i;
      directory *dir = f.path;
      if (dir == 0)
      {
         cerr << "Internal error 1: " << f.local << " has no path" << endl;
         return 1;
      }

      if (!dir->is_local_done() && match_local(dir, dirv, filev) == 0)
         dir->set_local_done();
   }
   return 0;
}

///////////////////////////////////////////////////////////////
// argument processing

int argument_option(const char *arg,
   file_opt& fopt, dir_v& dirv, ftp_link& ftp)
{
   if (arg == 0 || *arg == 0) return 0;

   char c;
   while ((c = *arg++) != 0)
   {
      switch(c)
      {
         case '-': fopt.set_minus(); break;
         case 'p': fopt.set_xfer(file_opt::put); break;
         case 'g': fopt.set_xfer(file_opt::get); break;
         case '2': fopt.set_xfer(file_opt::two_way); break;
         case '0': fopt.set_xfer(file_opt::no_way); break;
         case 't': fopt.set_date_set(); break;
         case 'n': fopt.reset_date_set(); break;
         case 'a': fopt.set_mode(file_opt::ascii); break;
         case 'b': fopt.set_mode(file_opt::binary); break;
         case 'i': fopt.set_upper(file_opt::ignore); break;
         case 's': fopt.set_upper(file_opt::sensitive); break;
         case 'r': fopt.set_upper(file_opt::rename); break;
         case 'L': return dirv.current_directory(arg) == 0 ? 1 : 0;
         case 'R': return ftp.current_directory(arg);
         case 'T':
         {
            long l;
            char *t = 0;
            l = strtoul(arg, &t, 10);
            if (t == 0 || *t != 0)
            {
               cerr << "Bad number " << arg << endl;
               return false;
            }

            ftp.set_time_bias(l);
            return true;
         }
         case 'H': return ftp.host(arg);
         case 'U': return ftp.user(arg);
         case 'P': return ftp.password(arg);
         case 'V': ++verbose; break;
         default:
         {
            if (c == 'h' || c == '?')
            {
               cout << copyright << endl;
               if (verbose == 0)
               {
                  cout << "(use -Vh for short help, -VVh for longer help)"
                     << endl;
               }
               else
               {
                  cout << help_msg << endl
                     << (verbose == 1 ? short_help_msg : long_help_msg)
                     << endl;
               }
            }
            else
               cerr << "Invalid option " << c << " (try -Vh)" << endl;
            return 1;
         }
      }
   }
   return 0;
}

int argument_file(const char *arg, file_opt& fopt, dir_v& dirv, file_v& filev)
{
   if (arg == 0 || *arg == 0) return 0;

   const char *argarg = arg;
   directory *prev = 0;

   string s;
   char c = *arg;
   bool is_relative;
   bool is_wild = false;
   bool has_space = false;

   if (c == '/' || c == '\\')
   {
      is_relative = false;
      ++arg;
   }
   else
   {
      is_relative = true;
   }

   while ((c = *arg++) != 0)
   {
      if (c == '/' || c == '\\')
      {
         if (is_wild)
         {
            cerr << "Cannot use wildcards on directory in " << argarg << endl;
            return 1;
         }
         if (s.empty())
         {
            cerr << "Bad path " << argarg << endl;
            return 1;
         }

         if (prev == 0)
         {
            if (is_relative)
               prev = dirv.current_directory();
            else
               prev = dirv.root_directory();
         }
         directory *cur = dirv.find_directory(s.c_str(), prev,
            &directory::local);
         if (cur == 0)
         {
            directory dir;
            dir.local = dir.remote = s;
            if (has_space)
            {
               remove_space(dir.remote);
               has_space = false;
            }

            dir.prev = prev;
            dirv.push_back(dir);
            cur = &dirv.back();
         }
         prev = cur;
         s.remove();
      }
      else if (c == ':')
      {
         if (s.size() != 1 || prev != 0 || is_wild)
         {
            cerr << "Invalid file name " << argarg << endl;
            return 1;
         }
         if (*arg == '/' || *arg == '\\')
         {
            is_relative = false;
            prev = dirv.root_directory(s[0]);
            ++arg;
         }
         else
            prev = dirv.current_directory(0, s[0]);
         s.remove();
      }
      else
      {
         if (c == '?' || c == '*')
            is_wild = true;
         else if (isspace(c))
            has_space = true;
         else if (c == '.' && s.size() == 0)
         {
            if (*arg == '/' || *arg == '\\')
            {
               ++arg;
               continue; // discard "xyz/./etc"
            }
            // other cases? (e.g. "xyz/." or "xyz/../etc")
         }
         s.append(1, c);
      }
   }

   if (s.size() == 0)
   {
      cerr << "Invalid filename " << argarg << endl;
      return 1;
   }

   if (prev == 0)
   {
      assert(is_relative);
      prev = dirv.current_directory();
   }

   file f;
   f.local = f.remote = s;
   f.opt = fopt;
   f.path = prev;
   if (is_wild)
   {
      f.opt.set_wild();
      if (has_space)
         remove_space(f.remote);
   }
   else
   {
      if (fopt.is_ignore())
         lower_nospace(f.remote);
      else if (has_space)
         remove_space(f.remote);

      // paranoid check: cannot specify same file twice
      file_v::iterator ckfit = filev.find_file(s.c_str(), prev, &file::local);
      if (ckfit != filev.end() && !(*ckfit).opt.is_wild())
      {
         cerr << "Repeated: " << (*ckfit).local
            << " already specified before " << argarg << endl;
         return 1;
      }
   }
   filev.push_back(f);

   return 0;
}

int argument(const char *arg, file_opt& fopt,
   dir_v& dirv, file_v& filev, ftp_link& ftp)
{
   if (arg == 0 || *arg == 0) return 0;

   if (arg[0] == '@')
   {
      filebuf fb;
      streambuf *pb;

      if (*++arg == 0)
         pb = cin.rdbuf();
      else
      {
         fb.open(arg, ios::in);
         if (!fb.is_open())
         {
            cerr << "Cannot open list file " << arg+1 << endl;
            return 1;
         }
         pb = &fb;
      }

      istream f(pb);
      string s;
      file_opt save_fopt = fopt;
      static const int ctrl_z = 26;
      while (f.good())
      {
         getline(f, s, '\n');
         f.clear(f.rdstate() &~ ios::failbit);
         if (s.size() && s[0] != '#')
         {
            istrstream ins((char*)s.c_str(), s.size());
            while (ins.good() && ins.peek() != ctrl_z)
            {
               string p;
               ins >> p;
               if (p.size() && argument(p.c_str(), fopt, dirv, filev, ftp))
                  return 1;
            }
         }
      }
      fopt = save_fopt;
      fopt.reset_minus();
      return 0;
   }

   if (arg[0] == '-')
   {
      char c = *++arg;
      if (c != '-' && c != '@')
         return argument_option(arg, fopt, dirv, ftp);
   }

   return argument_file(arg, fopt, dirv, filev);
}

/////////////////////////////////////
// control handler

BOOL WINAPI handler_routine(DWORD dwCtrlType)
{
  switch(dwCtrlType)
    {
    case CTRL_CLOSE_EVENT:
    case CTRL_LOGOFF_EVENT:
    case CTRL_SHUTDOWN_EVENT:
       // By returning TRUE, the user will be asked to terminate
       // the process and has a chance to let it go

       return (TRUE);

    case CTRL_C_EVENT:
    case CTRL_BREAK_EVENT:
    default:
       // don't handle the event
       return (FALSE);
    }
}




/////////////////////////////////////
// dialog for ftp password

// defines copied in from .rc file
#define IDD_FTPX          1
#define IDC_HOSTTEXT    101
#define IDC_HOST        102
#define IDC_USERTEXT    103
#define IDC_USER        104
#define IDC_PASSTEXT    105
#define IDC_PASSWORD    106

struct dlg_param
{
   string host, user, password;
};

extern "C" BOOL CALLBACK
dlg_proc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
   int wmId;

   switch (msg)
   {
      case WM_INITDIALOG:
      {
         // We need to initialize stuff in the dialog box...

         SetWindowLong(hdlg, GWL_USERDATA, lParam);
         dlg_param& p = *(dlg_param*)lParam;
         CenterWindow (hdlg);
         SetDlgItemText(hdlg, IDC_HOST, (LPCSTR)p.host.c_str());
         SetDlgItemText(hdlg, IDC_USER, (LPCSTR)p.user.c_str());

         // equivalent to SetDlgItemText
         HWND hpass = GetDlgItem(hdlg, IDC_PASSWORD);
         SendMessage(hpass, WM_SETTEXT, 0, (LPARAM)p.password.c_str());
         SetFocus(hpass);

         // explicitely show window (WS_VISIBLE not enough on NT?)
         ShowWindow(hdlg, SW_SHOW);
         UpdateWindow(hdlg);

         return FALSE; // dont set focus: I did it already
      }
      // would fall through
      case WM_DESTROY:
         break;

      case WM_COMMAND:
         wmId = LOWORD(wParam);
         switch (wmId)
         {
            case IDOK:
            {
               dlg_param& p = *(dlg_param*)GetWindowLong(hdlg, GWL_USERDATA);
               if (getdlgstring(IDC_HOST, p.host, hdlg) &&
                  getdlgstring(IDC_USER, p.user, hdlg) &&
                  getdlgstring(IDC_PASSWORD, p.password, hdlg))
               {
                  EndDialog(hdlg, 1);
                  return (TRUE);
               }
               else
                  return (FALSE);
            }
            case IDCANCEL:
               EndDialog(hdlg, 2);
               return (TRUE);
            default:
               break;
         }
      // fall through
      default:
         break;
   }
   return (FALSE);
}

int get_dlg_param(dlg_param *p)
{
   HINSTANCE hinst = GetModuleHandle (NULL);
   int ret = DialogBoxParam(hinst, MAKEINTRESOURCE(IDD_FTPX), 0,
      dlg_proc, (LPARAM)p);

   if (-1 == ret) return 0;
   return ret;
}


//////////////////////////////////////////////////////////////////////
// implementation of class functions

int directory::build_path(string& dest, path_mode pm) const
{
   unsigned int size = 0;
   int root_slash = 0;
   char drive;

   vector<const string*> path;

   for (const directory *i = this; i != 0; i = i->prev)
   {
      if (i->is_fail()) return 1; // recursive failure
      if (i->is_root())
      {
         root_slash = 1;
         drive = i->drive;
         assert(i->prev == 0);
      }
      else
      {
         const string& s = pm == local_path ? i->local : i->remote;
         if (s.empty()) // error: no name given (not root)
         {
            cerr << "Internal error: ";
            if (pm == local_path)
               cerr << "Empty local path for remote=" << i->remote << endl;
            else
               cerr << "Empty remote path for local=" << i->local << endl;
            return 1;
         }
         path.push_back(&s);
         size += s.size();
         if (i->is_current())
         {
            drive = i->drive;
            assert(i->prev == 0);
         }
      }
   }

   size += path.size() + root_slash;
   if (pm == local_path) size += 2;
   if (size > dest.max_size()) // definitely too long
   {
      cerr << "Internal error: ";
      cerr << "Reached size=" << size << " building path for ";
      if (pm == local_path) cerr << local; else cerr << remote;
      cerr << endl;
      return 1;
   }

   dest.remove();
   dest.reserve(size);
   const char slash = pm == local_path ? '\\' : '/';

   if (pm == local_path)
   {
      dest.append(1, drive);
      dest.append(1, ':');
   }

   if (root_slash != 0)
      dest.append(1, slash);

   for (int i = path.size() - 1; i >= 0; --i)
   {
      dest.append(*path[i]);
      dest.append(1, slash);
   }

   return 0;
}

void file::dump() const
{
   if (!opt.is_wild())
   {
      cout << "Local:  ";
      if (is_local_match())
         cout << local_ft << endl;
      else
         cout << "N/A" << endl;

      cout << "Remote: ";
      if (is_remote_match())
         cout << remote_ft << endl;
      else
         cout << "N/A" << endl;
   }

   // bug in rw lib? check against ios::right
   assert((cout.flags() & ios::adjustfield) != ios::right);

   if (verbose > 2)
   {
      cout << "local:  done=" << bool_YN(is_local_done())
         << "  match=" << bool_YN(is_local_match()) << endl;
      cout << "remote: done=" << bool_YN(is_remote_done())
         << "  match=" << bool_YN(is_remote_match()) << endl;
      cout << "fail=" << bool_YN(is_fail());
      if (opt.is_put()) cout << "  " << "put";
      if (opt.is_get()) cout << "  " << "get";
      if (opt.is_ascii()) cout << "  " << "ascii";
      if (opt.is_binary()) cout << "  " << "binary";
      if (opt.is_ignore()) cout << "  " << "ignore";
      if (opt.is_sensitive()) cout << "  " << "sensitive";
      if (opt.is_rename()) cout << "  " << "rename";
      if (opt.is_wild()) cout << "  " << "wild";
      cout << endl;
   }
   cout << endl;
}

directory *dir_v::root_directory(char drive)
{
   bool ok = get_drive(drive);
   for (iterator i = begin(); i != end(); ++i)
      if ((*i).is_root() && (*i).drive == drive)
         return &*i;

   directory dir;
   dir.drive = drive;
   dir.set_root();
   if (!ok) dir.set_fail(); // msg by get_drive
   push_back(dir);
   ++nroot;
   return &back();
}

directory *dir_v::current_directory(const char *p, char drive)
{
   // if a drive is given in the path, it must not be given twice
   assert(p == 0 || p[0] == 0 || p[1] != ':' || drive == 0);

   // build default arguments as needed
   char def_path[_MAX_PATH];
   if (p == 0 || *p == 0) // default path (with given or default drive)
   {
      p = 0;
      if (!get_drive(drive))
         return 0;
   }
   else if (p[1] == ':') // drive given in path
   {
      drive = *p;
      if (!get_drive(drive))
         return 0;
      p += 2;
   }
   else if (!get_drive(drive)) // default drive
      return 0;

   // check this is the only current path for given drive
   for (iterator i = begin(); i != end(); ++i)
      if ((*i).is_current() && (*i).drive == drive)
      {
         if ((*i).local.empty())
         {
            cerr << "Internal error 3: " << (*i).local
               << " has no path" << endl;
            return 0;
         }
         if (p != 0 && stricmp(p, (*i).local.c_str()) != 0)
         {
            cerr << "Multiple current paths for drive "
               << drive << ':' << (*i).local << " and "
               << p << endl;
            return 0;
         }
         return &*i;
      }

   // add a new current drive
   directory dir;
   dir.drive = drive;
   dir.local = p != 0 ? p : def_path;
   dir.set_current();
   if (p == 0) // default path
   {
      const int idrv = drive - 'A' + 1;
      if (_getdcwd(idrv, def_path, sizeof def_path) == 0 || def_path[0] == 0)
      {
         cerr << "Cannot getcwd" << endl;
         return 0;
      }
      p = def_path;
      if (p[0] != 0 && p[1] == ':') p += 2; // ??
   }
   dir.local = p;

   push_back(dir);
   ++ncurrent;
   return &back();
}

directory *dir_v::find_directory(const char *name, const directory *prev,
   const string directory::*member)
{
   for (iterator i = begin(); i != end(); ++i)
      if ((*i).prev == prev && stricmp(name, ((*i).*member).c_str()) == 0)
         return &*i;
   return 0;
}

file_v::iterator file_v::find_file(const char *name, const directory *path,
   const string file::*member)
{
   assert(name != 0);
   assert(path != 0);
   assert(member == &file::local || member == &file::remote);

   const bool is_local = member == &file::local;

   for (iterator i = begin(); i != end(); ++i)
      if ((*i).path == path)
      {
         bool ignore = is_local || (*i).opt.is_ignore(),
            wild = (*i).opt.is_wild();
         int (*tst)(const char*, const char*) =
            wild ? (ignore ? wildistr : wildstr) : (ignore ? stricmp : strcmp);
         if ((*tst)(((*i).*member).c_str(), name) == 0) return i;
      }
   return end();
}

bool ftp_link::connect() // true if successful
{
   assert(!is_connect());
   assert(m_host.size() > 0);

   if (m_user.empty()) m_user = "anonymous";
   if (m_pass.empty())
   {
      dlg_param p;
      p.user = m_user;
      p.host = m_host;
      if (get_dlg_param(&p) != 1) // not OK
      {
         report_error("Remote connection cancelled by user");
         return false;
      }
      m_host = p.host;
      m_user = p.user;
      m_pass = p.password;
   }

   if (verbose > 1)
      cout << "Starting FTP connection to " << m_host
         << " user: " << m_user << endl;

   hopen = InternetOpen("ftpx", INTERNET_OPEN_TYPE_DIRECT, 0, 0, 0);
   if (hopen == 0)
   {
      report_error("InternetOpen failed");
      return false;
   }

   hconn = InternetConnect(hopen, m_host.c_str(),
      INTERNET_DEFAULT_FTP_PORT, m_user.c_str(), m_pass.c_str(),
      INTERNET_SERVICE_FTP, 0, 0);

   if (hconn == 0)
   {
      report_error("Cannot open FTP to ",
         m_host.c_str(), " user: ", m_user.c_str());
      InternetCloseHandle(hopen);
      hopen = 0;
      return false;
   }

   return connected = true;
}

bool ftp_link::get_current_directory(string& s)
{
   if (m_directory.size() > 0)
   {
      s = m_directory;
      return true;
   }

   if (is_connect())
   {
      char buf[512];
      DWORD size = sizeof buf;
      if (FtpGetCurrentDirectory(hconn, buf, &size))
         m_directory = buf;
      else
      {
         char *altbuf = 0;
         if (GetLastError() == ERROR_INSUFFICIENT_BUFFER &&
            (altbuf = (char*)malloc(size)) != 0 &&
            FtpGetCurrentDirectory(hconn, altbuf, &size))
               m_directory = altbuf;
         if (altbuf != 0) free(altbuf);
      }

      if (m_directory.size() > 0)
      {
         if (verbose > 1)
            cout << "ftp: PWD is " << m_directory << endl;
         s = m_directory;
         return true;
      }

      report_error("Cannot get current ftp directory");
   }
   return false;
}

bool ftp_link::set_working_directory(const string& s)
{
   assert(is_connect());

   if (s == CWD) return true;

   bool rtc = FtpSetCurrentDirectory(get_handle(), s.c_str());
   if (rtc) CWD = s;
   return rtc;
}

bool ftp_link::disconnect()
{
   assert(is_connect());
   bool rtc = true;

   if (verbose > 1)
      cout << "Disconnecting from " << m_host << endl;

   if (!InternetCloseHandle(hconn))
   {
      rtc = false;
      report_error("Closing FTP conection");
   }
   if (!InternetCloseHandle(hopen))
   {
      rtc = false;
      report_error("Closing Internet");
   }

   hopen = hconn = 0;
   connected = false;

   assert(!is_connect());
   return rtc;
}

bool ftp_link::put_file(const string& local, const string& remote, mode m)
{
   assert(is_connect());

   bool rtc = FALSE != FtpPutFile(hconn, local.c_str(), remote.c_str(),
      INTERNET_FLAG_DONT_CACHE |
      (m == ascii ? FTP_TRANSFER_TYPE_ASCII : FTP_TRANSFER_TYPE_BINARY), 0);
   if (!rtc)
      report_error("Cannot put local file ", local.c_str(),
         " to ", remote.c_str());
   return rtc;
}

bool ftp_link::get_file(const string& local, const string& remote, mode m)
{
   assert(is_connect());

   bool rtc = FALSE != FtpGetFile(hconn, remote.c_str(), local.c_str(), FALSE,
      FILE_ATTRIBUTE_NORMAL, INTERNET_FLAG_DONT_CACHE |
      (m == ascii ? FTP_TRANSFER_TYPE_ASCII : FTP_TRANSFER_TYPE_BINARY), 0);
   if (!rtc)
      report_error("Cannot get local file ", local.c_str(),
         " from ", remote.c_str());
   return rtc;
}

//////////////////////////////////////////////////////////////////////
// main function

int main_func(int argc, char *argv[])
{
   cerr = get_errbuf(!isatty(fileno(stdout)));

   dir_v dirv;
   file_v filev;
   ftp_link ftp;

   // default options
   file_opt fopt(file_opt::two_way, file_opt::ascii, file_opt::ignore);

   // read arguments
   for (int i = 1; i < argc; ++i)
      if (argument(argv[i], fopt, dirv, filev, ftp))
         return 1;

   // add current directories (any drive's current matches remote current)
   if (ftp.is_named())
   {
      string rcur;
      if (!ftp.connect() || !ftp.get_current_directory(rcur))
         return 1;
      int ncurrent = 0;
      for (dir_v::iterator i = dirv.begin(); i != dirv.end(); ++i)
         if ((*i).is_current())
         {
            ++ncurrent;
            (*i).remote = rcur;
         }
      if (ncurrent == 0)
      {
         directory *cur = dirv.current_directory();
         if (cur == 0) return 1;

         cur->remote = rcur;
      }
   }

   // locate files and complete file data
   if (add_local_files(dirv, filev))
      return 1;
   if (ftp.is_connect() && add_remote_files(dirv, filev, ftp))
      return 1;

   // sort files
   if (verbose > 1)
      cout << "Sorting " << filev.size() << " file(s) ..." << flush;
   filev.sort();
   if (verbose > 1) cout << endl;


   // perform put/get/list/rename as needed
   // get correct file/time for transferred files (marked 'done')
   int nget = 0, nput = 0, nok = 0, nrename = 0, nerrs = 0;
   bool do_further = false;
   file_v::iterator redo = filev.end();
   int nredo = 0;
   string rpath, lpath;
   directory *cur = 0;
   for (file_v::iterator i = filev.begin(); i != filev.end(); ++i)
   {
      file& f = *i;

      // track remote directory (files are sorted on remote path)
      assert(f.path != 0);
      if (f.path != cur)
      {
         cur = track_remote(redo, nredo, f.path, lpath, rpath, filev, ftp);
         redo = filev.end();
         nredo = 0;
         if (cur == 0) continue;
      }

      // compare file date/time and decide action, rename, further
      // (-1=get, 0=equal, 1=put)
      int compare = CompareFileTime(&f.local_ft, &f.remote_ft);
      // 0= equal, 1= get, 2= put
      if (compare < 0) compare = 1; else if (compare > 0) compare = 2;
      int action;
      if (f.is_fail() || f.opt.is_wild())
         action = 0; // no action
      else if (f.is_local_match() && f.is_remote_match())
      {
         if (compare == 0)
            action = 1; // no op
         else if (compare == 1 && f.opt.is_get() ||
            compare == 2 && f.opt.is_put())
               action = 2; // do as compare directs
         else action = 1; // inhibited by option
      }
      else if (!ftp.is_connect())
         action = 1; // no op
      else if (f.is_local_match() && compare == 2 && f.opt.is_put() ||
         f.is_remote_match() && compare == 1 && f.opt.is_get())
            action = 2; // do as compare directs
      else if (f.is_local_match() || f.is_remote_match())
         action = 1; // inhibited by option
      else
         action = 0; // no match


      // do screen output
      if (verbose > 0 || action == 2)
      {
         static const char * const status[3][3] =
         //  equal   get     put
          { {" ?? ", " ?? ", " ?? "},          // action == 0
            {" == ", " !< ", " >! "},          // action == 1
            {" ?? ", " << ", " >> "}};         // action == 2

         if (verbose > 1 || action > 0)
         {
            cout << lpath << f.local;
            if (ftp.is_connect())
               cout << status[action][compare] << rpath << f.remote;
            cout << endl;
         }
         if (verbose > 1) f.dump();
      }


      // perform any file copying, set further rename action
      if (action == 1)
      {
         if (compare == 0) ++nok;
         action = 3;
      }
      else if (action == 2)
      {
         ftp_link::mode ftp_mode = f.opt.is_ascii() ? ftp.ascii : ftp.binary;
         if (compare == 1) // get
         {
            if (ftp.get_file(lpath + f.local, f.remote, ftp_mode))
            {
               ++nget;
               // on gotten files: redate anyway!
               f.set_local_done();
               do_further = true;
               cur->set_further();
               action = 3;
            }
            else
               ++nerrs;
         }
         else if (compare == 2) // put
         {
            if (ftp.put_file(lpath + f.local, f.remote, ftp_mode))
            {
               ++nput;
               if (f.opt.is_date_set())
               {
                  if (redo == filev.end()) redo = i;
                  f.set_remote_done();
                  ++nredo;
                  do_further = true;
                  cur->set_further();
               }
               action = 3;
            }
            else
               ++nerrs;
         }
      }


      // rename files
      if (f.opt.is_rename() && action == 3 &&
         strcmp(f.local.c_str(), f.remote.c_str()) != 0)
      {
         string oldname = lpath + f.local, newname = lpath + f.remote;
         if (verbose > 2)
            cout << "Renaming " << oldname << " as " << newname << endl;
         if (MoveFile(oldname.c_str(), newname.c_str()) == FALSE)
         {
            ++nerrs;
            cerr << "Cannot rename " << oldname << " as " << newname << endl;
         }
         else
         {
            ++nrename;
            f.local = f.remote;
         }
      }
   }
   if (nredo > 0) // get dates for last directory, if needed
      track_remote(redo, nredo, 0, lpath, rpath, filev, ftp);

   if (ftp.is_connect())
      ftp.disconnect();

   // set remote dates for copied files
   redo = filev.end();
   int nredate = 0;
   if (do_further)
   {
      cur = 0;
      for (file_v::iterator i = filev.begin(); i != filev.end(); ++i)
      {
         file& f = *i;
         if (!f.path->is_further())
            continue;

         if (f.path != cur)
         {
            cur = track_remote(redo, 0, f.path, lpath, rpath, filev, ftp);
            if (cur == 0) continue;
         }

         if (f.is_local_done()) // change date as remote
         {
            const string s = lpath + f.local;
            if (set_file_time(s.c_str(), &f.remote_ft) == 0)
               ++nredate;
            else
               ++nerrs;
         }
         else if (f.is_remote_done())
         {
            cerr << "File " << lpath << f.local
               << " was put but not found" << endl;
         }
      }
   }

   if (verbose > 0)
   {
      long fl = cout.flags();
      cout.setf(ios::right, ios::adjustfield);
      if (nget != 0 || verbose > 1)
         cout << setw(9) << nget << setw(0) << " file(s) gotten\n";
      if (nput != 0 || verbose > 1)
         cout << setw(9) << nput << setw(0) << " file(s) put\n";
      if (nok != 0 || verbose > 1)
         cout << setw(9) << nok << setw(0) << " file(s) found equal\n";
      if (nrename != 0 || verbose > 1)
         cout << setw(9) << nrename << setw(0) << " file(s) renamed\n";
      if (nredate != 0 || verbose > 1)
         cout << setw(9) << nredate << setw(0) << " file(s) redated\n";
      cout << setw(9) << nerrs << setw(0) << " error(s)" << endl;
      cout.flags(fl);
   }
   return 0;
}

int main(int argc, char *argv[])
{
   SetConsoleCtrlHandler(handler_routine, TRUE);
   int rtc = main_func(argc, argv);
   SetConsoleCtrlHandler(handler_routine, FALSE);
   return rtc;
}
